@@ -19,16 +19,16 @@
 #endif
 
 #include "time_zone.h"
-#include "time_zone_if.h"
 
 #include <cctype>
 #include <chrono>
+#include <cstddef>
 #include <cstdint>
 #include <cstring>
 #include <ctime>
 #include <limits>
+#include <string>
 #include <vector>
-
 #if !HAS_STRPTIME
 #include <iomanip>
 #include <sstream>
@@ -37,10 +37,15 @@
 #endif
 #endif
 
+#include "civil_time.h"
+#include "time_zone_if.h"
+
 namespace cctz {
 namespace detail {
-#if !HAS_STRPTIME
 
+namespace {
+
+#if !HAS_STRPTIME
 // Build a strptime() using C++11's std::get_time().
 inline char* strptime(const char* s, const char* fmt, std::tm* tm) {
   std::istringstream input(s);
@@ -50,7 +55,6 @@ inline char* strptime(const char* s, const char* fmt, std::tm* tm) {
 #else
   input >> std::get_time(tm, fmt);
 #endif
-
   if (input.fail()) return nullptr;
   
   if (input.tellg() > 0) {
@@ -59,16 +63,8 @@ inline char* strptime(const char* s, const char* fmt, std::tm* tm) {
       return const_cast<char*>(s) + strlen(s);
   }
 }
-
-#else
-
-inline char* strptime(const char* s, const char* fmt, std::tm* tm) {
-    return ::strptime(s, fmt, tm);
-}
-
 #endif
 
-namespace {
 std::tm ToTM(const time_zone::absolute_lookup& al) {
   std::tm tm{};
   tm.tm_sec = al.cs.second();
@@ -119,14 +115,14 @@ const char kDigits[] = "0123456789";
 // Formats a 64-bit integer in the given field width.  Note that it is up
 // to the caller of Format64() [and Format02d()/FormatOffset()] to ensure
 // that there is sufficient space before ep to hold the conversion.
-char* Format64(char* ep, int width, std::int64_t v) {
+char* Format64(char* ep, int width, std::int_fast64_t v) {
   bool neg = false;
   if (v < 0) {
     --width;
     neg = true;
-    if (v == INT64_MIN) {
-      // Avoid negating INT64_MIN.
-      int last_digit = -(v % 10);
+    if (v == std::numeric_limits<std::int_fast64_t>::min()) {
+      // Avoid negating minimum value.
+      std::int_fast64_t last_digit = -(v % 10);
       v /= 10;
       if (last_digit < 0) {
         ++v;
@@ -174,10 +170,10 @@ void FormatTM(std::string* out, const std::string& fmt, const std::tm& tm) {
   // an error, like the array wasn't large enough.  To accomodate this,
   // the following code grows the buffer size from 2x the format string
   // length up to 32x.
-  for (int i = 2; i != 32; i *= 2) {
-    size_t buf_size = fmt.size() * i;
+  for (std::size_t i = 2; i != 32; i *= 2) {
+    std::size_t buf_size = fmt.size() * i;
     std::vector<char> buf(buf_size);
-    if (size_t len = strftime(&buf[0], buf_size, fmt.c_str(), &tm)) {
+    if (std::size_t len = strftime(&buf[0], buf_size, fmt.c_str(), &tm)) {
       out->append(&buf[0], len);
       return;
     }
@@ -236,11 +232,12 @@ const char* ParseInt(const char* dp, int width, T min, T max, T* vp) {
   return dp;
 }
 
-// The number of base-10 digits that can be represented by an std::int64_t.
-const int kDigits10_64 = std::numeric_limits<std::int64_t>::digits10;
+// The number of base-10 digits that can be represented by a signed 64-bit
+// integer.  That is, 10^kDigits10_64 <= 2^63 - 1 < 10^(kDigits10_64 + 1).
+const int kDigits10_64 = 18;
 
-// 10^n for everything that can be represented by an std::int64_t.
-const std::int64_t kExp10[kDigits10_64 + 1] = {
+// 10^n for everything that can be represented by a signed 64-bit integer.
+const std::int_fast64_t kExp10[kDigits10_64 + 1] = {
     1,
     10,
     100,
@@ -281,9 +278,12 @@ const std::int64_t kExp10[kDigits10_64 + 1] = {
 //
 // We also handle the %z and %Z specifiers to accommodate platforms that do
 // not support the tm_gmtoff and tm_zone extensions to std::tm.
+//
+// Requires that zero() <= fs < seconds(1).
 std::string format(const std::string& format, const time_point<sys_seconds>& tp,
-                   const std::chrono::nanoseconds& ns, const time_zone& tz) {
+                   const detail::femtoseconds& fs, const time_zone& tz) {
   std::string result;
+  result.reserve(format.size());  // A reasonable guess for the result size.
   const time_zone::absolute_lookup al = tz.lookup(tp);
   const std::tm tm = ToTM(al);
 
@@ -308,7 +308,7 @@ std::string format(const std::string& format, const time_point<sys_seconds>& tp,
 
     // If the new pending text is all ordinary, copy it out.
     if (cur != start && pending == start) {
-      result.append(pending, cur - pending);
+      result.append(pending, static_cast<std::size_t>(cur - pending));
       pending = start = cur;
     }
 
@@ -319,7 +319,7 @@ std::string format(const std::string& format, const time_point<sys_seconds>& tp,
     // If the new pending text is all percents, copy out one
     // percent for every matched pair, then skip those pairs.
     if (cur != start && pending == start) {
-      size_t escaped = (cur - pending) / 2;
+      std::size_t escaped = static_cast<std::size_t>(cur - pending) / 2;
       result.append(pending, escaped);
       pending += escaped * 2;
       // Also copy out a single trailing percent.
@@ -338,43 +338,43 @@ std::string format(const std::string& format, const time_point<sys_seconds>& tp,
       }
       switch (*cur) {
         case 'Y':
-          // This avoids the tm_year overflow problem for %Y, however
+          // This avoids the tm.tm_year overflow problem for %Y, however
           // tm.tm_year will still be used by other specifiers like %D.
           bp = Format64(ep, 0, al.cs.year());
-          result.append(bp, ep - bp);
+          result.append(bp, static_cast<std::size_t>(ep - bp));
           break;
         case 'm':
           bp = Format02d(ep, al.cs.month());
-          result.append(bp, ep - bp);
+          result.append(bp, static_cast<std::size_t>(ep - bp));
           break;
         case 'd':
         case 'e':
           bp = Format02d(ep, al.cs.day());
           if (*cur == 'e' && *bp == '0') *bp = ' ';  // for Windows
-          result.append(bp, ep - bp);
+          result.append(bp, static_cast<std::size_t>(ep - bp));
           break;
         case 'H':
           bp = Format02d(ep, al.cs.hour());
-          result.append(bp, ep - bp);
+          result.append(bp, static_cast<std::size_t>(ep - bp));
           break;
         case 'M':
           bp = Format02d(ep, al.cs.minute());
-          result.append(bp, ep - bp);
+          result.append(bp, static_cast<std::size_t>(ep - bp));
           break;
         case 'S':
           bp = Format02d(ep, al.cs.second());
-          result.append(bp, ep - bp);
+          result.append(bp, static_cast<std::size_t>(ep - bp));
           break;
         case 'z':
           bp = FormatOffset(ep, al.offset / 60, '\0');
-          result.append(bp, ep - bp);
+          result.append(bp, static_cast<std::size_t>(ep - bp));
           break;
         case 'Z':
           result.append(al.abbr);
           break;
         case 's':
           bp = Format64(ep, 0, ToUnixSeconds(tp));
-          result.append(bp, ep - bp);
+          result.append(bp, static_cast<std::size_t>(ep - bp));
           break;
       }
       pending = ++cur;
@@ -391,7 +391,7 @@ std::string format(const std::string& format, const time_point<sys_seconds>& tp,
         FormatTM(&result, std::string(pending, cur - 2), tm);
       }
       bp = FormatOffset(ep, al.offset / 60, ':');
-      result.append(bp, ep - bp);
+      result.append(bp, static_cast<std::size_t>(ep - bp));
       pending = ++cur;
     } else if (*cur == '*' && cur + 1 != end &&
                (*(cur + 1) == 'S' || *(cur + 1) == 'f')) {
@@ -400,7 +400,7 @@ std::string format(const std::string& format, const time_point<sys_seconds>& tp,
         FormatTM(&result, std::string(pending, cur - 2), tm);
       }
       char* cp = ep;
-      bp = Format64(cp, 9, ns.count());
+      bp = Format64(cp, 15, fs.count());
       while (cp != bp && cp[-1] == '0') --cp;
       switch (*(cur + 1)) {
         case 'S':
@@ -411,7 +411,7 @@ std::string format(const std::string& format, const time_point<sys_seconds>& tp,
           if (cp == bp) *--bp = '0';
           break;
       }
-      result.append(bp, cp - bp);
+      result.append(bp, static_cast<std::size_t>(cp - bp));
       pending = cur += 2;
     } else if (*cur == '4' && cur + 1 != end && *(cur + 1) == 'Y') {
       // Formats %E4Y.
@@ -419,7 +419,7 @@ std::string format(const std::string& format, const time_point<sys_seconds>& tp,
         FormatTM(&result, std::string(pending, cur - 2), tm);
       }
       bp = Format64(ep, 4, al.cs.year());
-      result.append(bp, ep - bp);
+      result.append(bp, static_cast<std::size_t>(ep - bp));
       pending = cur += 2;
     } else if (std::isdigit(*cur)) {
       // Possibly found %E#S or %E#f.
@@ -433,12 +433,12 @@ std::string format(const std::string& format, const time_point<sys_seconds>& tp,
           bp = ep;
           if (n > 0) {
             if (n > kDigits10_64) n = kDigits10_64;
-            bp = Format64(bp, n, (n > 9) ? ns.count() * kExp10[n - 9]
-                                         : ns.count() / kExp10[9 - n]);
+            bp = Format64(bp, n, (n > 15) ? fs.count() * kExp10[n - 15]
+                                          : fs.count() / kExp10[15 - n]);
             if (*np == 'S') *--bp = '.';
           }
           if (*np == 'S') bp = Format02d(bp, al.cs.second());
-          result.append(bp, ep - bp);
+          result.append(bp, static_cast<std::size_t>(ep - bp));
           pending = cur = ++np;
         }
       }
@@ -488,16 +488,15 @@ const char* ParseZone(const char* dp, std::string* zone) {
   return dp;
 }
 
-const char* ParseSubSeconds(const char* dp,
-                            std::chrono::nanoseconds* subseconds) {
+const char* ParseSubSeconds(const char* dp, detail::femtoseconds* subseconds) {
   if (dp != nullptr) {
-    std::int64_t v = 0;
-    std::int64_t exp = 0;
+    std::int_fast64_t v = 0;
+    std::int_fast64_t exp = 0;
     const char* const bp = dp;
     while (const char* cp = strchr(kDigits, *dp)) {
       int d = static_cast<int>(cp - kDigits);
       if (d >= 10) break;
-      if (exp < 9) {
+      if (exp < 15) {
         exp += 1;
         v *= 10;
         v += d;
@@ -505,8 +504,8 @@ const char* ParseSubSeconds(const char* dp,
       ++dp;
     }
     if (dp != bp) {
-      v *= kExp10[9 - exp];
-      *subseconds = std::chrono::nanoseconds(v);
+      v *= kExp10[15 - exp];
+      *subseconds = detail::femtoseconds(v);
     } else {
       dp = nullptr;
     }
@@ -517,7 +516,7 @@ const char* ParseSubSeconds(const char* dp,
 // Parses a string into a std::tm using strptime(3).
 const char* ParseTM(const char* dp, const char* fmt, std::tm* tm) {
   if (dp != nullptr) {
-    dp = cctz::detail::strptime(dp, fmt, tm);
+    dp = strptime(dp, fmt, tm);
   }
   return dp;
 }
@@ -537,21 +536,21 @@ const char* ParseTM(const char* dp, const char* fmt, std::tm* tm) {
 //
 // We also handle the %z specifier to accommodate platforms that do not
 // support the tm_gmtoff extension to std::tm.  %Z is parsed but ignored.
-//
-// TODO: Support parsing year values beyond the width to tm_year.
 bool parse(const std::string& format, const std::string& input,
            const time_zone& tz, time_point<sys_seconds>* sec,
-           std::chrono::nanoseconds* ns) {
+           detail::femtoseconds* fs, std::string* err) {
   // The unparsed input.
   const char* data = input.c_str();  // NUL terminated
 
   // Skips leading whitespace.
   while (std::isspace(*data)) ++data;
 
-  const int kintmax = std::numeric_limits<int>::max();
-  const int kintmin = std::numeric_limits<int>::min();
+  const year_t kyearmax = std::numeric_limits<year_t>::max();
+  const year_t kyearmin = std::numeric_limits<year_t>::min();
 
   // Sets default values for unspecified fields.
+  bool saw_year = false;
+  year_t year = 1970;
   std::tm tm{};
   tm.tm_year = 1970 - 1900;
   tm.tm_mon = 1 - 1;  // Jan
@@ -562,8 +561,9 @@ bool parse(const std::string& format, const std::string& input,
   tm.tm_wday = 4;  // Thu
   tm.tm_yday = 0;
   tm.tm_isdst = 0;
-  auto subseconds = std::chrono::nanoseconds(0);
-  int offset = kintmin;
+  auto subseconds = detail::femtoseconds::zero();
+  bool saw_offset = false;
+  int offset = 0;  // No offset from passed tz.
   std::string zone = "UTC";
 
   const char* fmt = format.c_str();  // NUL terminated
@@ -571,7 +571,7 @@ bool parse(const std::string& format, const std::string& input,
   bool afternoon = false;
 
   bool saw_percent_s = false;
-  std::int64_t percent_s_time = 0;
+  std::int_fast64_t percent_s = 0;
 
   // Steps through format, one specifier at a time.
   while (data != nullptr && *fmt != '\0') {
@@ -598,11 +598,11 @@ bool parse(const std::string& format, const std::string& input,
     }
     switch (*fmt++) {
       case 'Y':
-        // We're more liberal than the 4-digit year typically handled by
-        // strptime(), but we still need to store the result in an int,
-        // and the intermediate value has a 1900 excess.
-        data = ParseInt(data, 0, kintmin + 1900, kintmax, &tm.tm_year);
-        if (data != nullptr) tm.tm_year -= 1900;
+        // Symmetrically with FormatTime(), directly handing %Y avoids the
+        // tm.tm_year overflow problem.  However, tm.tm_year will still be
+        // used by other specifiers like %D.
+        data = ParseInt(data, 0, kyearmin, kyearmax, &year);
+        if (data != nullptr) saw_year = true;
         continue;
       case 'm':
         data = ParseInt(data, 2, 1, 12, &tm.tm_mon);
@@ -634,12 +634,16 @@ bool parse(const std::string& format, const std::string& input,
         break;
       case 'z':
         data = ParseOffset(data, '\0', &offset);
+        if (data != nullptr) saw_offset = true;
         continue;
       case 'Z':  // ignored; zone abbreviations are ambiguous
         data = ParseZone(data, &zone);
         continue;
       case 's':
-        data = ParseInt(data, 0, INT64_MIN, INT64_MAX, &percent_s_time);
+        data = ParseInt(data, 0,
+                        std::numeric_limits<std::int_fast64_t>::min(),
+                        std::numeric_limits<std::int_fast64_t>::max(),
+                        &percent_s);
         if (data != nullptr) saw_percent_s = true;
         continue;
       case 'E':
@@ -650,19 +654,20 @@ bool parse(const std::string& format, const std::string& input,
           } else {
             data = ParseOffset(data, ':', &offset);
           }
+          if (data != nullptr) saw_offset = true;
           fmt += 1;
           continue;
         }
         if (*fmt == '*' && *(fmt + 1) == 'S') {
           data = ParseInt(data, 2, 0, 60, &tm.tm_sec);
-          if (data != NULL && *data == '.') {
+          if (data != nullptr && *data == '.') {
             data = ParseSubSeconds(data + 1, &subseconds);
           }
           fmt += 2;
           continue;
         }
         if (*fmt == '*' && *(fmt + 1) == 'f') {
-          if (data != NULL && std::isdigit(*data)) {
+          if (data != nullptr && std::isdigit(*data)) {
             data = ParseSubSeconds(data, &subseconds);
           }
           fmt += 2;
@@ -670,10 +675,10 @@ bool parse(const std::string& format, const std::string& input,
         }
         if (*fmt == '4' && *(fmt + 1) == 'Y') {
           const char* bp = data;
-          data = ParseInt(data, 4, -999, 9999, &tm.tm_year);
+          data = ParseInt(data, 4, year_t{-999}, year_t{9999}, &year);
           if (data != nullptr) {
             if (data - bp == 4) {
-              tm.tm_year -= 1900;
+              saw_year = true;
             } else {
               data = nullptr;  // stopped too soon
             }
@@ -686,14 +691,14 @@ bool parse(const std::string& format, const std::string& input,
           if (const char* np = ParseInt(fmt, 0, 0, 1024, &n)) {
             if (*np == 'S') {
               data = ParseInt(data, 2, 0, 60, &tm.tm_sec);
-              if (data != NULL && *data == '.') {
+              if (data != nullptr && *data == '.') {
                 data = ParseSubSeconds(data + 1, &subseconds);
               }
               fmt = ++np;
               continue;
             }
             if (*np == 'f') {
-              if (data != NULL && std::isdigit(*data)) {
+              if (data != nullptr && std::isdigit(*data)) {
                 data = ParseSubSeconds(data, &subseconds);
               }
               fmt = ++np;
@@ -714,7 +719,7 @@ bool parse(const std::string& format, const std::string& input,
 
     // Parses the current specifier.
     const char* orig_data = data;
-    std::string spec(percent, fmt - percent);
+    std::string spec(percent, static_cast<std::size_t>(fmt - percent));
     data = ParseTM(data, spec.c_str(), &tm);
 
     // If we successfully parsed %p we need to remember whether the result
@@ -722,7 +727,8 @@ bool parse(const std::string& format, const std::string& input,
     // So reparse the input with a known AM hour, and check if it is shifted
     // to a PM hour.
     if (spec == "%p" && data != nullptr) {
-      std::string test_input = "1" + std::string(orig_data, data - orig_data);
+      std::string test_input = "1";
+      test_input.append(orig_data, static_cast<std::size_t>(data - orig_data));
       const char* test_data = test_input.c_str();
       std::tm tmp{};
       ParseTM(test_data, "%I%p", &tmp);
@@ -735,59 +741,87 @@ bool parse(const std::string& format, const std::string& input,
     tm.tm_hour += 12;
   }
 
-  if (data == nullptr) return false;
+  if (data == nullptr) {
+    if (err != nullptr) *err = "Failed to parse input";
+    return false;
+  }
 
   // Skip any remaining whitespace.
   while (std::isspace(*data)) ++data;
 
   // parse() must consume the entire input string.
-  if (*data != '\0') return false;
+  if (*data != '\0') {
+    if (err != nullptr) *err = "Illegal trailing data in input string";
+    return false;
+  }
 
   // If we saw %s then we ignore anything else and return that time.
   if (saw_percent_s) {
-    *sec = FromUnixSeconds(percent_s_time);
-    *ns = std::chrono::nanoseconds::zero();
+    *sec = FromUnixSeconds(percent_s);
+    *fs = detail::femtoseconds::zero();
     return true;
   }
 
   // If we saw %z or %Ez then we want to interpret the parsed fields in
   // UTC and then shift by that offset.  Otherwise we want to interpret
   // the fields directly in the passed time_zone.
-  time_zone ptz = tz;
-  if (offset != kintmin) {
-    ptz = utc_time_zone();  // Override tz.  Offset applied later.
-  } else {
-    offset = 0;  // No offset from passed tz.
-  }
+  time_zone ptz = saw_offset ? utc_time_zone() : tz;
 
   // Allows a leap second of 60 to normalize forward to the following ":00".
   if (tm.tm_sec == 60) {
     tm.tm_sec -= 1;
     offset -= 1;
-    subseconds = std::chrono::nanoseconds::zero();
+    subseconds = detail::femtoseconds::zero();
   }
 
-  std::int64_t year = tm.tm_year;
-  if (year > INT64_MAX - 1900) {
-    year = INT64_MAX;
-  } else {
+  if (!saw_year) {
+    year = year_t{tm.tm_year};
+    if (year > kyearmax - 1900) {
+      // Platform-dependent, maybe unreachable.
+      if (err != nullptr) *err = "Out-of-range year";
+      return false;
+    }
     year += 1900;
   }
 
-  // TODO: Eliminate extra normalization.
-  const civil_second cs(year, tm.tm_mon + 1, tm.tm_mday,
-                        tm.tm_hour, tm.tm_min, tm.tm_sec);
+  const int month = tm.tm_mon + 1;
+  civil_second cs(year, month, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);
 
-  // parse() fails if any normalization was done.  That is,
-  // parsing "Sep 31" will not produce the equivalent of "Oct 1".
-  if (cs.year() != year || cs.month() != tm.tm_mon + 1 ||
-      cs.day() != tm.tm_mday || cs.hour() != tm.tm_hour ||
-      cs.minute() != tm.tm_min || cs.second() != tm.tm_sec) {
+  // parse() should not allow normalization. Due to the restricted field
+  // ranges above (see ParseInt()), the only possibility is for days to roll
+  // into months. That is, parsing "Sep 31" should not produce "Oct 1".
+  if (cs.month() != month || cs.day() != tm.tm_mday) {
+    if (err != nullptr) *err = "Out-of-range field";
     return false;
   }
 
-  *sec = ptz.lookup(cs).pre - sys_seconds(offset);
-  *ns = subseconds;
+  // Accounts for the offset adjustment before converting to absolute time.
+  if ((offset < 0 && cs > civil_second::max() + offset) ||
+      (offset > 0 && cs < civil_second::min() + offset)) {
+    if (err != nullptr) *err = "Out-of-range field";
+    return false;
+  }
+  cs -= offset;
+
+  const auto tp = ptz.lookup(cs).pre;
+  // Checks for overflow/underflow and returns an error as necessary.
+  if (tp == time_point<sys_seconds>::max()) {
+    const auto al = ptz.lookup(time_point<sys_seconds>::max());
+    if (cs > al.cs) {
+      if (err != nullptr) *err = "Out-of-range field";
+      return false;
+    }
+  }
+  if (tp == time_point<sys_seconds>::min()) {
+    const auto al = ptz.lookup(time_point<sys_seconds>::min());
+    if (cs < al.cs) {
+      if (err != nullptr) *err = "Out-of-range field";
+      return false;
+    }
+  }
+
+  *sec = tp;
+  *fs = subseconds;
   return true;
 }
 
